---
elm:
  dependencies:
    avh4/elm-fifo: latest
    elm/browser: latest
    elm/core: latest
    elm/html: latest
    elm/json: latest
    elm/parser: latest
    elm/regex: latest
    elm/time: latest
    elm/url: latest
    elm/virtual-dom: latest
    elm-community/graph: latest
    elm-community/intdict: latest
    elm-community/list-extra: latest
    elm-community/maybe-extra: latest
    elm-community/string-extra: latest

  source-directories:
    - ../src
---

<!-- ---
elm:
  dependencies:
    vicramgon/logicus: latest
--- -->

# Sintáxis y Semántica LP en LogicUS

## Objetivos

Para el desarrollo de los aspectos prácticos de la asignatura, que extenderán naturalmente a los aspectos teórico-formales estudiados en la mismas, utilizaremos el lenguaje Elm, de origen declarativo pero con orientación a la interación Web, en concreto con las lobrerías LogicUS, que nos permitirán, a lo largo del curso, trasladar a un entorno más visual cada uno de los conceptos y algoritmos vistos en la asignatura.

En esta primera práctica veremos cómo trasladar los aspectos teórico formales de la sintáxis y semántica proposicionales en LogicUS abordando:

- La sintáxis de fórmulas (y sus árboles de formación) y conjuntos proposicionales.

- Definición de interpretaciones y evaluación de fórmulas y conjuntos respecto de las interpretaciones.

- Satisfactibilidad y validez lógica, modelos y contramodelos, tablas de verdad y consecuencia lógica.

## 1. Sintáxis de la Lógica Proposicional en LogicUS

### 1.1. Sintáxis de fórmulas

Si recordamos, la definición de las fórmulas proposicionalesn se ha dado como definición recursiva a partir de átomos (variables proposicionales) bajo la aplicación de las conectivas lógicas (negación, conjunción, disyunción, implicación y equivalencia). Siguiendo dicha definición constructiva se establece en LogicUS la estructura de las fórmulas de forma que:

- Los símbolos proposicionales se definen como cadenas de caractéres (con la restricción de que deben usar minúsculas):

        type alias PSymb = String

- Las fórmulas proposicionales se definen según la estructura recursiva dada por:

        type FormulaLP = Atom PSymb                 -- caso base
                        | Neg FormulaLP             --⎫
                        | Conj FormulaLP FormulaLP  --⎪
                        | Disj FormulaLP FormulaLP  --⎬ casos recursivos
                        | Impl FormulaLP FormulaLP  --⎪
                        | Equi FormulaLP FormulaLP  --⎭
                        | Insat                     -- fórmula insatisfactible

De forma que se tiene como elemento fundamental el átomo y a partir de éste y con la aplicación de las conectivas a partir de subfórmulas (según la aridad de la conectiva) se permite la definición de cualquier fórmula proposicional. Además, se añade una representación explícita para la fórmula insatisfactible ($\perp$).

Bien, una vez vista la clara analogía entre el formalismo y la implementación, pasemos directamente a definir las fórmulas. Para ello es necesario importar los módulos necesarios con los que vayamos a trabajar, en este caso bastará importar **LogicUS.PL.SintaxSemantics**.

> **_Nota_**: _En el [repositorio de paquetes de Elm](https://package.elm-lang.org/) se encuentran los paquetes publicados. En ellos pueden consultarse cómodamete las funciones que se encuentran definidas en cada uno de sus módulos. En concreto para el paquete [LogicUS](https://package.elm-lang.org/packages/vicramgon/logicus/latest/)._

Para importar el módulo (ya se da declarado en las dependdencias) es suficiente con ejecutar (en este caso en un chunk de litvis)

```elm {l id=imports}
import LogicUS.PL.SintaxSemantics exposing (..)
```

Con _exposing_ se pueden especificar qué funciones y tipos se quieren importar específicamente pero por el momento importaremos todas ellas con _(..)_.

Una vez importados tenemos varias vías para definir las funciones:

##### A) Haciendo uso de los constructores

Para ello basta con utilizar los constructores (_Atom, Neg, Conj, Disj, Impl, Equi, Insat_). Para ello, a una variable (el identificador debe comenzar por una letra minúscula) se le asocia la fórmula descrita de forma recursiva de forma que:

- Para definir un átomo con la variable "p" basta escrbir:

      f = Atom "p"

- Para definir la negación de una fórmula, por ejemplo del átomo anterior basta escribir:

      f2 = Neg f

  o equivalentemente

      f2 = Neg (Atom "p")

- Para definir la conjunción, disyunción, implicación o equivalencia es necesario dar como argumento dos fórmulas. Por ejemplo, si consideramos la fórmula $p \wedge \neg p$ entonces debemos escribir:

      f3 = Conj f f2

  o equivalentemente

      f3 = Conj (Atom "p") (Neg (Atom "p"))

Hagamos algunas puntualizaciones:

1. Aunque hemos dicho que realizar _f2 = Neg f_ era equivalente a _f2 = Neg (Atom "p")_ , en realidad no lo es. Elm es un lenguaje funcional y por tanto todos los objetos del lenguaje son funciones. De manera que cuando establecemos _Neg f_, _f_ no es más que una referencia a la función _f_ de forma que si cambia _f_ entonces consecuentemente cambiará también _f2_, mientras que en el segundo caso al haberse definido a partir de elementos constantes (constructores y cadenas) entonces "f2" sólo cambiará si se modifica explícitamente.

2. Nótese que para todos los constructores el formato es similar _Identificador Costructor + parámetros_. De forma que los parámetros se dan según la aridad del mismo separados por espacios y encerrados entre paréntesis si no constan de un único elemento.

Según lo expuesto intetemos definir algunas funciones.

---

> **_EJEMPLO_**: Formalizar los siguientes enunciados, definiendo las fórmulas correspondientes:
>
> 1.  Si el sol brilla hoy, entonces no brillará mañana. ($p$: el sol brilla hoy, $q$:el sol brillará mañana)
>
> 2.  Llueve o nieva siempre que la presión atmosférica está baja y sólo entonces. ($p$: llueve, $q$:nieva, $r$:la presión está baja)
>
> 3.  Si Pablo no se encontró con Chari ayer, entonces no tomaron café juntos ni pasearon por el parque. ($p$: Pablo se encontró con Chari, $q$: Pablo y chari tomaron café juntos, $r$: Plablo y Chari pasearon por el parque)

- Apartado 1:

  ```elm {l context=1 follows=imports}
  f1 = Impl (Atom "p") (Neg (Atom "q"))
  ```

  ^^^elm {m=("$ F_1 =" ++ (fplToMathString f1) ++ "$") context=1}^^^

- Apartado 2:

  ```elm {l context=1}
  f2 = Equi (Disj (Atom "p") (Atom "q")) (Atom "r")
  ```

  ^^^elm {m=("$ F_2 =" ++ (fplToMathString f2) ++ "$") context=1}^^^

- Apartado 3:

  ```elm {l context=1}
  f3 = Impl (Neg (Atom "p")) (Conj (Neg (Atom "q")) (Neg (Atom "r")))
  ```

  ^^^elm {m=("$ F_3 =" ++ (fplToMathString f3) ++ "$") context=1}^^^

---

##### B) Con lectura desde una cadena

Desde luego, escribir las fórmulas a través de los constructores puede convertirse en una tarea ardua y propensa a errores, ya que es necesario construir previamente el árbol de formación y reproducirlo en el lenguaje. Por eso se proporciona una función de parseo que es capaz de leerlas de una cadena. Para que esa cadena pueda ser interpretada correctamente es necesario que cumpla las siguentes condiciones:

- Símbolos proposicionales: según la expresión regular: $$\texttt{[a-z]+(\_\{[0-9]+(,[0-9]+)*})?\}$$

  Esto es, una serie de caracteres en minúscula segidos (opcionalmente) por una lísta de subíndices especificados entre llaves.

- Conectivas. Siguiendo la correspondencia y prioridad de asociación:

$$
\def\arraystretch{1.5} \begin{array}{|c|c|}
\hline
\textbf{\small Conectiva Lógica} & \textbf{\small Símbolo LogicUS} \\ \hline
\textrm{\small Negación} & \neg \\
\hline
\textrm{\small Conjunción} & /\backslash \\
\hline
\textrm{\small Disyunción} & \backslash/ \\
\hline
\textrm{\small Implicación} & -> \\
\hline
\textrm{\small Equivalencia} & <-> \\
\hline
\end{array}
$$

- Uso de paréntesis para cambiar prioridad y establecer asociaciones. No son necesarios los paréntesis externos.

Las funciones que permiten realizar la lectura y extraerla son _fplReadFromString_ y _fplReadExtraction_. La primera nos devolverá un error si no puede leer la fórmula.

---

> **_EJEMPLO_**: Formalizar los siguientes enunciados, definiendo las fórmulas correspondientes:
>
> 1.  Si el sol brilla hoy, entonces no brillará mañana. ($p$: el sol brilla hoy, $q$:el sol brillará mañana)
>
> 2.  Llueve o nieva siempre que la presión atmosférica está baja y sólo entonces. ($p$: llueve, $q$:nieva, $r$:la presión está baja)
>
> 3.  Si Pablo no se encontró con Chari ayer, entonces no tomaron café juntos ni pasearon por el parque. ($p$: Pablo se encontró con Chari, $q$: Pablo y chari tomaron café juntos, $r$: Plablo y Chari pasearon por el parque)

- Apartado 1:

  ```elm {l context=2 follows=imports}
  f1 = fplReadExtraction <| fplReadFromString "p -> ¬ q"
  ```

  ^^^elm {m=("$ F_1 =" ++ (fplToMathString f1) ++ "$") context=2}^^^

- Apartado 2:

  ```elm {l context=2}
  f2 = fplReadExtraction <| fplReadFromString "p | q <-> r"
  ```

  ^^^elm {m=("$ F_2 =" ++ (fplToMathString f2) ++ "$") context=2}^^^

  _Nota:_ Veáse que el carácter "\\" denota un símbolo de escape por lo que para poner la barra invertida es necesario escribir "\\\\".

  <br>

- Apartado 3:

  ```elm {l context=2}
  f3 = fplReadExtraction <| fplReadFromString "¬p -> ¬q & ¬r"
  ```

  ^^^elm {m=("$$ F_3 =" ++ (fplToMathString f3) ++ "$$") context=2}^^^

  Si hubiésemos querido hacer uso de subíndices, en este último ejemeplo, podríamos haber escrito:

  ```elm {l context=2}
  f4 = fplReadExtraction <| fplReadFromString "¬p_{1} -> ¬q_{1} & ¬r_{1}"
  ```

  ^^^elm {m=("$" ++ (fplToMathString f4) ++ "$") context=2}^^^

Para mostrar las fórmulas podemos utilizar la función _fplToString_ o la función _fplToMathString_ (en formato latex), utilzando los _triple hat chuncks_.

---

### 1.2. Árboles de formación

En teoría se ha visto la representación de fórmulas mediante árboles de formación que contienen cada una de sus subfórmulas. La libería también permite también extraer los árboles de formación de las fórmulas, en concreto, a través de las funciones _fplFormTree_ y _fplFormTreeToDOT_. La primera es la que computa el árbol mientras que la segunda da una representación del mismo en formato DOT que puede ser renderizado en un visualizador GraphViz.

> ###### VISUALIZACIÓN DE DOT EN LITVIS
>
> Litvis posee dicho visualizador en un chunk tipo dot en el cuál podemos pegar la dot String generada. De forma que para visualizar cualquier gráfico DOT generado por la libería podemos hacer los isiguiente:
>
> 1.  Crear una variable que la almacene definiendo su tipo como String (_var : String_).
>
> 2.  En el chunk establecer los parámetros _l_ (para mostrar el chunk) y _m_ para ejecutar la salida como markdown lo nos mostrará el código en formato DOT. (Es indispensable establecer el tipo, si no se pone no se mostrará nada).
>
> 3.  El código DOT mostrado podemos copiarlo en un chunk _dot_ para visualizarlo.
>
> 4.  Para dejar de mostrar el código DOT basta con quitar el parámetro _m_ del chunk.

Por ejemplo si queremos extraer el árbol de formación de la última fórmula que hemos definido: $F_3 \equiv \neg p \rightarrow \neg q \wedge \neg r$,

```elm {l context=2}
ft_f3 : String
ft_f3 = fplFormTreeToDOT <| fplFormTree f3
```

<center>

```dot
digraph G {
rankdir=TB
graph []
node [shape=plaintext, color=black]
edge [dir=none]

0 -> 1
0 -> 3
1 -> 2
3 -> 4
3 -> 6
4 -> 5
6 -> 7

0 [label="( ¬ p → ( ¬ q ∧ ¬ r ) )"]
1 [label="¬ p"]
2 [label="p"]
3 [label="( ¬ q ∧ ¬ r )"]
4 [label="¬ q"]
5 [label="q"]
6 [label="¬ r"]
7 [label="r"]
}
```

</center>

### 1.3. Definición de conjuntos de fórmulas

Visto cómo podemos definir fórmulas, ¿cómo podemos ahora definir conjuntos de fórmulas?. En LogicUS los conjuntos de fórmulas están definidos como listas, de forma que para definir conjuntos es suficiente con `definir` las fórmulas y después el conjunto como una lista de ellas.

    type alias LPSet = List FormulaLP

De forma que si, por ejemplo, queremos definir el conjunto:

$ M = \left\lbrace (p \wedge q) \vee (p \wedge r), (p \wedge r) \vee (\neg p \wedge q) \rightarrow \neg q, (p \leftrightarrow q) \wedge (p \rightarrow \neg q) \wedge p \right\rbrace$

Bastaría escribir:

```elm {l context=3 follows=imports}
f1 = fplReadExtraction <| fplReadFromString "p & q | p & r"
f2 = fplReadExtraction <| fplReadFromString " p & r | ¬ p & q -> ¬q"
f3 = fplReadExtraction <| fplReadFromString "(p <-> q) & (q -> p) & p "

m = [f1, f2, f3]
```

Podemos mostrarlo en formato latex utilizando la función _splToMathString_:

^^^elm {m=("$" ++ (splToMathString m) ++ "$") context=3}^^^

## 2. Semántica de la Lógica Proposicional en LogicUS

### 2.1. Interpretaciones

Como se ha estudiado en la teoría, los **valores de verdad** corresponden a 1 (_verdadero_) y 0 (_falso_). Elm ya provee esos valores booleanos en el tipo _Bool_, por lo que no es necesario realizar ninguna definición alternativa para este concepto.

La definición de las funciones de verdad asociadas a las conectivas corresponden directamente a la aplicación de dichas funciones en la evaluación de las fórmulas. Antes de ver la evaluación resulta necesario ver la definición que se ha dado para las interpretaciones. Para ello, se ha elegido una representación "dispersa" que indica por medio de una lista los símbolos proposicionales de las variables que son tomadas como verdaderas. Las variables proposicionales asociadas a símbolos que no aparecen en la lista serán considerados como falsas.

    type alias Interpretation = List PSymb

Con esta definición resulta sencillo llevar a cabo la evaluación de las fórmulas siguiendo exactamente los dos pasos anotados:

- **Evaluación de variables**. Una variable será verdadera si pertenece a la lista de la interpretación, y será falsa en caso contrario.

- **Evaluación de conectivas**. Aplicando las correspondientes funciones características:

      fplValuation : FormulaLP -> Interpretation -> Bool
      fplValuation f i =
          case f of
              Atom symb -> List.member symb i
              Neg g -> not (valuation g i)
              Conj g h -> valuation g i &&  valuation h i
              Disj g h -> valuation g i ||  valuation h i
              Impl g h -> not (valuation g i) || valuation h i
              Equi g h -> (valuation g &&  valuation h i) || ( not(valuation g i) &&  not (valuation h i))
              Insat -> False

  NOTA: nótese que la fórmula insatisfactible es falsa respecto de cualquier valoración.

Luego, para realizar la evaluación de una fórmula respecto de una interpretación podemos utilizar la función _fplValuation_ dando la fórmula y la interpretación. De forma que si queremos hallar la valoración de $F \equiv \neg(\neg(p \vee q)\vee(\neg r \vee s))$ respecto de $i \equiv \{p=1, q=1, r=0, s=0\}$ bastaría definir $f$ y $i$ y ejecutar la función _fplValuation f i_.

```elm {l context=4 follows=imports}
f = fplReadExtraction <| fplReadFromString "¬ (¬ (p | q) | (¬ r | s))"

i = ["p", "q"]
```

^^^elm{m=(fplValuation f i) context=4}^^^

De igual modo, se puede hallar la valoración de un conjunto respecto de una valoración. Recuérdese que la valoración de un conjunto respecto a una valoración era verdadera si y sólo si todas las fórmulas del conjunto eran verdaderas respecto de dicha valoración. La función _splValuation_ realiza justamente esta comparativa:

    splValuation : PLSet -> Interpretation -> Bool
    splValuation fs i =
        List.all (\x -> fplValuation x i) fs

De forma que, análogamente a lo presentado para las fórmulas, bastaría definir el conjunto, la interpretación y ejecutar la función _splValuation_.

### 2.2 Tablas de verdad, modelos y contramodelos

#### Tablas de verdad

Como se ha estudiado en el correspondiente apartado teórico, una tabla de verdad recoge las valoraciones de una fórmula respecto a todas las posibles interpretaciones de sus símbolos proposicionales. Nótese que dada la implementación propuesta para las interpretaciones, las posibles interpretaciones de una fórmula corresponderían a todos los posibles subconjuntos del conjunto de los símbolos proposicionales de la fórmula (un conjunto de $n$ elementos tiene $2^n$ subconjuntos distintos, que justamente coincide con el número de entradas de la tabla de verdad).

En la libería LogicUS existen funciones para calcular los símbolos e interpretaciones, pero pasemos directamente a ver la tabla de verdad, que puede ser calculada mediante la función _fplTruthTable_ de forma que dada una fórmula devuelve un conjunto (lista) de tuplas _(Interpretation, Bool)_ correspondientes a las entradas de la tabla y las correspondientes valoraciones. Además para poderla visualizar cómodamente, se presenta también la función _fplTruthTableMathString_ que recibe una fórmula y devuelve una cadena con la tabla en formato Latex.

De forma que, si queremos construir la tabla de la fórmula anterior $F \equiv \neg(\neg(p \vee q)\vee(\neg r \vee s))$ bastaría ejecutar _fplTruthTableMathString f_.

^^^elm{m=("$$" ++ (fplTruthTableMathString f) ++ "$$") context=4}^^^

De igual modo, para los conjuntos existen funciones de cálculo y representación de la tabla de verdad: _splTruthTable_ y _splTruthTableMathString_ , respectivamente.

#### Modelos y contramodelos

A partir de la tabla de verdad (o directamente de la evaluación de las posibles interpretaciones) es posible extraer los modelos y contramodelos de una fórmula. Recuérdese que los modelos correspondían a las interpretaciones que hacían a la fórmula verdadera y los contramodelos los que la hacían falsa. En LogicUS las funciones _fplModels_ y _fplCounterModels_ realizan precisamente ese filtrado de forma que dada una fórmula devuelven una lista con las interpretaciones modelo y una lista con los contramodelos respectivamente.

De foma que si queremos obtener los modelos de la fórmula $F \equiv \neg(\neg(p \vee q)\vee(\neg r \vee s))$ bastaría ejecutar _fplModels f_, dando como resultado:

^^^elm{m=("$" ++ (interpretationsToMathString (fplModels f) ["p", "q", "r", "s"])  ++ "$") context=4}^^^

O _fplCountermodels f_ para los contramodelos:

^^^elm{m=("$" ++ (interpretationsToMathString (fplCountermodels f) ["p", "q", "r", "s"]) ++ "$") context=4}^^^

De forma equivalente para los conjuntos se proporcionan las funciones _splModels_ y _splCounterModels_

### 2.3. Satisfactibilidad, validez y consecuencia lógica

Una vez visto el cálculo de modelos, y dado que la satisfactibilidad y validez lógicas se dan de forma que si todas las interpretaciones hacen a la fórmula verdadera se tiene _tautología_, si alguna de ellas la hace verdadera se dice _satisfactible_, o si ninguna la hace se dice _insatisfactible_. De forma que dadas todas las interpretaciones bastaría comprobar si para todas ellas son modelos, si alguna lo es o si ninguna lo es.

Eso es justamente lo que hacen las funciones _fplValidity_, _fplSatisfiablity_ y _fplUnsatisfiability_, de forma que dada una fórmula devuelven si es tautología (válida), satisfactible e insatisfactible, respectivamente. Por ejemplo, la fórmula $F$ es satisfactible y no tautología.

- ¿$F \in SAT$?: ^^^elm{r=(fplSatisfiability f) context=4}^^^
- ¿$F \in TAUT$?: ^^^elm{r=(fplValidity f) context=4}^^^

De igual forma se definen para los conjuntos las funciones _fplSatisfiability_ y _splUnsatisfiability_ para comprobar la consistencia de un conjunto. Y además se define la consecuencia lógica en la función _logicalConsecuence_ y _logicalConsecuence2_ que recibiendo un conjunto y una fórmula determinan si $U \models F$, la primera de ellas comparando si todo modelo de $U$ lo es también de $F$ y la segunda comprobando si $U \cup \{\neg F\} \in INSAT$.

De forma que, si por ejemplo queremos comprobar si $\{p \rightarrow q,q \rightarrow p \wedge r\} \models p \rightarrow (p \rightarrow q) \rightarrow r$ entonces:

```elm {l context=5 follows=imports}
f1 = fplReadExtraction <| fplReadFromString "p -> q"
f2 = fplReadExtraction <| fplReadFromString "q -> p & r"
f3 = fplReadExtraction <| fplReadFromString "p -> (p -> q) -> r"
```

- ¿$\{F_1, F_2\} \models F_3$? ^^^elm{r=(logicalConsecuence [f1,f2] f3) context=5}^^^ (_logicalConsecuence [f1,f2] f3_)
- ¿$\{F_1, F_2\} \models F_3$? ^^^elm{r=(logicalConsecuence2 [f1,f2] f3) context=5}^^^ (_logicalConsecuence2 [f1,f2] f3_)

<br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br>

